from dataclasses import dataclass

from serde import deserialize
from chris.types import (
    FeedId,
    Username,
    FilesUrl,
    FeedUrl,
    PluginInstancesUrl,
    NoteUrl,
    UserUrl,
    TagsUrl,
    TaggingsUrl,
    CommentsUrl,
)

from chris.helpers.connected_resource import ConnectedResource
from typing import List
from datetime import datetime


@deserialize
@dataclass(frozen=True)
class Note(ConnectedResource):
    url: NoteUrl
    id: int
    title: str
    content: str
    feed: FeedUrl


@deserialize
@dataclass(frozen=True)
class Feed(ConnectedResource):
    """
    A *feed* in *ChRIS* is a DAG of *plugin instances*.
    """

    url: FeedUrl
    id: FeedId
    creation_date: datetime
    modification_date: datetime
    name: str
    creator_username: Username
    created_jobs: int
    waiting_jobs: int
    scheduled_jobs: int
    started_jobs: int
    registering_jobs: int
    finished_jobs: int
    errored_jobs: int
    cancelled_jobs: int
    owner: List[UserUrl]
    note: NoteUrl
    tags: TagsUrl
    taggings: TaggingsUrl
    comments: CommentsUrl
    files: FilesUrl
    plugin_instances: PluginInstancesUrl

    def set_name(self, name: str) -> "Feed":
        res = self.session.put(self.url, data={"name": name})
        res.raise_for_status()
        return self.__refresh()

    def set_description(self, description: str) -> Note:
        res = self.session.put(
            self.note, data={"title": "Description", "content": description}
        )
        return Note.deserialize(res, self.session)

    def get_note(self) -> Note:
        return Note.deserialize(self.session.get(self.note), self.session)

    def __refresh(self) -> "Feed":
        res = self.session.get(self.url)
        return self.deserialize(res, self.session)



from typing import NewType, Union
from enum import Enum

# pyserde doesn't support typing.Literal

CUBEToken = NewType("CUBEToken", str)
CUBEAddress = NewType("CUBEAddress", str)
Username = NewType("Username", str)

PluginUrl = NewType("PluginUrl", str)
PluginId = NewType("PluginId", int)
PluginName = NewType("PluginName", str)
PluginVersion = NewType("PluginVersion", str)


SwiftPath = NewType("SwiftPath", str)


CUBEErrorCode = NewType("CUBEErrorCode", str)

ContainerImageTag = NewType("ContainerImageTag", str)

FeedId = NewType("FeedId", int)
PipingId = NewType("PipingId", int)
PipelineId = NewType("PipelineId", int)


ParameterName = NewType("ParameterName", str)
ParameterType = Union[str, int, float, bool]

PipelineParameterId = NewType("ParameterLocalId", int)
PluginParameterId = NewType("ParameterGlobalId", int)
PluginInstanceId = NewType("PluginInstanceId", int)

ComputeResourceName = NewType("ComputeResourceName", str)
FileResourceName = NewType("FileResourceName", str)
FileId = NewType("FileId", int)

UserUrl = NewType("UserUrl", str)
FilesUrl = NewType("FilesUrl", str)
FileResourceUrl = NewType("FileResourceUrl", str)
PipelineUrl = NewType("PipelineUrl", str)
PipingsUrl = NewType("PipingsUrl", str)
PipelinePluginsUrl = NewType("PipelinePluginsUrl", str)
PipelineDefaultParametersUrl = NewType("PipelineDefaultParametersUrl", str)
PipingUrl = NewType("PipingUrl", str)

PipelineParameterUrl = NewType("PipingParameterUrl", str)
PluginInstanceUrl = NewType("PluginInstanceUrl", str)
PluginInstancesUrl = NewType("PluginInstancesUrl", str)
DescendantsUrl = NewType("DescendantsUrl", str)
PipelineInstancesUrl = NewType("PipelineInstancesUrl", str)
PluginInstanceParamtersUrl = NewType("PluginInstanceParametersUrl", str)
ComputeResourceUrl = NewType("ComputeResourceUrl", str)
SplitsUrl = NewType("SplitsUrl", str)
FeedUrl = NewType("FeedUrl", str)
NoteUrl = NewType("NoteUrl", str)
"""A feed's note."""
PluginParametersUrl = NewType("PluginParametersUrl", str)
TagsUrl = NewType("TagsUrl", str)
TaggingsUrl = NewType("TaggingsUrl", str)
CommentsUrl = NewType("CommentsUrl", str)


class PluginType(Enum):
    ds = "ds"
    fs = "fs"
    ts = "ts"


class PluginInstanceStatus(Enum):
    created = "created"
    waiting = "waiting"
    scheduled = "scheduled"
    started = "started"
    registeringFiles = "registeringFiles"
    finishedSuccessfully = "finishedSuccessfully"
    finishedWithError = "finishedWithError"
    cancelled = "cancelled"


class ParameterTypeName(Enum):
    string = "string"
    integer = "integer"
    float = "float"
    boolean = "boolean"



from pathlib import Path
from dataclasses import dataclass, field
import requests
from requests import Session
from typing import Optional, Union
import functools

from chris.types import (
    CUBEAddress,
    CUBEToken,
    Username,
    PluginInstanceId,
    PluginUrl,
    PluginName,
    PluginVersion,
    PluginInstanceUrl,
)
from chris.cube.plugin import Plugin
from chris.cube.plugin_instance import PluginInstance
from chris.cube.files import File
from chris.cube.pipeline import Pipeline
from chris.helpers.pagination import Paginated
from chris.errors import ChrisIncorrectLoginError

from chris.helpers.peek import peek
from chris.cube.base_collection_links import BaseResponse, BaseCollectionLinks
from chris.helpers.deserialize import deserialize


@dataclass(frozen=True)
class ChrisClient:

    url: CUBEAddress
    token: CUBEToken

    collection_links: BaseCollectionLinks = field(init=False)
    _s: Session = field(init=False)

    def __post_init__(self):
        object.__setattr__(self, "_s", self.__authenticated_session(self.token))
        object.__setattr__(self, "collection_links", self.__get_collection_links())

    @classmethod
    def from_login(
        cls,
        address: Union[CUBEAddress, str],
        username: Union[Username, str],
        password: str,
    ) -> "ChrisClient":
        if not address.endswith("/api/v1/"):
            raise ValueError('Address of CUBE must end with "/api/v1/"')
        login = requests.post(
            address + "auth-token/", json={"username": username, "password": password}
        )
        if login.status_code == 400:
            res = login.json()
            raise ChrisIncorrectLoginError(
                res["non_field_errors"][0] if "non_field_errors" in res else login.text
            )
        login.raise_for_status()
        return cls(url=address, token=login.json()["token"])

    @staticmethod
    def __authenticated_session(token: CUBEToken) -> requests.Session:
        s = requests.Session()
        s.headers.update(
            {
                "Accept": "application/json",
                "Authorization": "Token " + token,
            }
        )
        return s

    def __get_collection_links(self) -> BaseCollectionLinks:
        """
        Make a request to the CUBE address. Calling this method verifies
        that the login token is correct.
        """
        res = self._s.get(self.url)
        if res.status_code == 401:
            raise ChrisIncorrectLoginError(res.json()["detail"])
        return deserialize(BaseResponse, res).collection_links

    def upload(self, local_file: Path, upload_path: str) -> File:
        """
        Upload a local file into *ChRIS*.

        Paramters
        ---------
        local_file: Path
            file to upload
        upload_path: str
            name of upload destination in *ChRIS* `uploadedfiles` resource
        """
        if not upload_path.startswith(f"{self.username}/uploads/"):
            upload_path = f"{self.username}/uploads/{upload_path}"

        with local_file.open("rb") as file_object:
            files = {
                "upload_path": (None, upload_path),
                "fname": (local_file.name, file_object),
            }
            res = self._s.post(
                self.collection_links.uploadedfiles,
                files=files,
            )

        return File.deserialize(res, self._s)

    def get_plugin_by_url(self, url: Union[PluginUrl, str]) -> Plugin:
        return Plugin.deserialize(self._s.get(url), self._s)

    def get_plugin_by_name(
        self,
        name_exact: Union[PluginName, str],
        version: Optional[Union[PluginVersion, str]] = None,
    ) -> Optional[Plugin]:
        search = self.search_plugin(name_exact, version)
        return peek(search)

    def search_plugin(
        self, name_exact: str, version: Optional[str] = None
    ) -> Paginated[Plugin]:
        qs = self._join_qs(name_exact=name_exact, version=version)
        return Paginated(
            item=Plugin,
            url=f"{self.collection_links.plugins}search/?{qs}",
            session=self._s,
        )

    def get_plugin_instance(
        self, plugin: Union[PluginInstanceUrl, PluginInstanceId, int, str]
    ) -> PluginInstance:
        """
        Get a plugin instance.
        :param plugin: Either a plugin instance ID or URL
        :return: plugin instance
        """
        if isinstance(plugin, str) and "/" in plugin:
            url = plugin
        else:
            url = f"{self.url}plugins/instances/{plugin}/"
        res = self._s.get(url)
        return PluginInstance.deserialize(res, self._s)

    def search_files(self, fname="", fname_exact="") -> Paginated[File]:
        return self._search_files(self.collection_links.files, fname, fname_exact)

    def search_uploadedfiles(self, fname="", fname_exact="") -> Paginated[File]:
        return self._search_files(
            self.collection_links.uploadedfiles, fname, fname_exact
        )

    def _search_files(
        self, base_url: str = "", fname: str = "", fname_exact: str = ""
    ) -> Paginated[File]:
        # TODO hard-coded feed files URL
        # https://github.com/FNNDSC/chrs/blob/fd7fdf997c5d751c8a95d510e70bdc5fd0a73f9a/chrs/src/download.rs#L44-L60
        if not base_url:
            base_url = self.collection_links.files
        qs = self._join_qs(fname=fname, fname_exact=fname_exact)
        url = f"{base_url}search/?{qs}"
        return self.get_files(url)

    def get_files(self, url: str) -> Paginated[File]:
        return Paginated(item=File, session=self._s, url=url)

    def get_file(self, fname_exact: str) -> Optional[File]:
        search = self._search_files(fname_exact=fname_exact)
        return peek(search)

    def search_pipelines(self, name="") -> Paginated[Pipeline]:
        return Paginated(
            item=Pipeline,
            url=f"{self.collection_links.pipelines}search/?name={name}",
            session=self._s,
        )

    def get_pipeline(self, name: str) -> Optional[Pipeline]:
        return peek(self.search_pipelines(name))

    @property
    def username(self) -> Username:
        return Username(self.__user["username"])

    @functools.cached_property
    def __user(self) -> dict:
        res = self._s.get(url=self.collection_links.user)
        res.raise_for_status()
        return res.json()

    @staticmethod
    def _join_qs(**kwargs) -> str:
        return "&".join([f"{k}={v}" for k, v in kwargs.items() if v])



from dataclasses import dataclass
from serde import deserialize

from chris.types import UserUrl, FilesUrl, TagsUrl


@deserialize()
@dataclass(frozen=True)
class BaseCollectionLinks:
    chrisinstance: str
    admin: str
    files: FilesUrl
    compute_resources: str
    plugin_metas: str
    plugins: str
    plugin_instances: str
    pipelines: str
    pipeline_instances: str
    tags: TagsUrl
    uploadedfiles: FilesUrl
    pacsfiles: FilesUrl
    servicefiles: FilesUrl
    filebrowser: str
    user: UserUrl


@deserialize()
@dataclass(frozen=True)
class BaseResponse:
    collection_links: BaseCollectionLinks



from dataclasses import dataclass
from typing import Optional

from chris.types import (
    PluginUrl,
    PipelineId,
    ParameterName,
    ParameterType,
    ParameterTypeName,
    PipingId,
    PipelineParameterId,
    PluginParameterId,
    PluginName,
    PluginVersion,
    PluginId,
    PipingUrl,
    PluginParametersUrl,
    PipelineParameterUrl,
    PipelineUrl,
)
from chris.helpers.connected_resource import ConnectedResource
from serde import deserialize


@deserialize()
@dataclass(frozen=True)
class PipelineParameter(ConnectedResource):
    url: PipelineParameterUrl
    id: PipelineParameterId
    value: ParameterType
    type: ParameterTypeName
    plugin_piping_id: PipingId
    previous_plugin_piping_id: None
    param_name: ParameterName
    param_id: PluginParameterId
    plugin_piping: PipingUrl
    plugin_name: PluginName
    plugin_version: PluginVersion
    plugin_id: PluginId
    plugin_param: PluginParametersUrl


@deserialize()
@dataclass(frozen=True)
class Piping(ConnectedResource):
    url: PipingUrl
    id: PipingId
    plugin_id: PluginId
    pipeline_id: PipelineId
    previous: Optional[PipingUrl]
    plugin: PluginUrl
    pipeline: PipelineUrl
    previous_id: Optional[PipingId] = None



from os import PathLike
from dataclasses import dataclass
from contextlib import contextmanager
from typing import Iterable, Sized, Iterator, Union, ContextManager
from requests import Response

from chris.types import FileResourceUrl, FileResourceName
import serde
from chris.helpers.connected_resource import ConnectedResource


@dataclass(frozen=True)
class Stream(Iterable[bytes], Sized):
    """
    Byte-stream.
    """

    fsize: int
    chunk_size: int
    res: Response

    def __iter__(self) -> Iterator[bytes]:
        return self.res.iter_content(chunk_size=self.chunk_size)

    def __len__(self) -> int:
        return self.fsize


@serde.deserialize()
@dataclass(frozen=True)
class File(ConnectedResource):
    """
    A file in _ChRIS_.
    """

    url: str
    fname: FileResourceName
    fsize: int
    file_resource: FileResourceUrl

    def download(self, destination: Union[PathLike, str], chunk_size=8192):
        """
        Download this file to a path.
        """
        with self.stream(chunk_size) as stream:
            with open(destination, "wb") as f:
                for data in stream:
                    f.write(data)

    @contextmanager
    def stream(self, chunk_size: int) -> ContextManager[Stream]:
        """
        Download the file as a bytes-stream, useful for adding a progress bar to large files.

        # Example

        ```python
        from tqdm import tqdm

        with open("file.dat", "wb") as f:
            with file.stream(8196) as stream:
                for chunk in tqdm(stream):
                    f.write(chunk)
        ```
        """
        with self.session.get(
            self.file_resource, stream=True, headers={"Accept": None}
        ) as r:
            r.raise_for_status()
            yield Stream(self.fsize, chunk_size, r)
